#include <stdio.h>
#include <string.h>

static int destkorgm = 1;


static const char* korgmname = 0;
static char* korg9name = 0;
static FILE* korgm = 0;
static FILE* korg9 = 0;

const char* KorgFileID()
{
  if (destkorgm)
    return "KORGM";
  return "KORGI";
}

static long readlong(FILE* input)
{
  long v = 0;
  for (int i = 0; i < 32; i+=8)
  {
    int c = fgetc(input);
    v += ((long)c)<< i;
  }
  return v;
}

static int readword(FILE* input)
{
  int v = 0;
  for (int i = 0; i < 16; i+=8)
  {
    int c = fgetc(input);
    v += c << i;
  }
  return v;
}

static void writeword(int v, FILE* output)
{
  fputc(v & 0xff, output);
  fputc((v >>8) & 0xff, output);
}

static long copybytes(FILE* input, FILE* output, long size)
{
  long n = 0;
  while (size-- > 0)
  {
   fputc(fgetc(input), output);
   n++;
  }
  return n;
}

static void copystyle(char* name, long ofs, long size)
{
  printf("copying style %10.10s ...\n", name);
  fseek(korg9, ofs+10, SEEK_SET);
  size -= 10; // without duplicated name

  //long newstylepos = ftell (korgm);
  fwrite(name, 10, 1, korgm);
  fwrite("      ", 6, 1, korgm); // names are 16 bytes long
  fwrite("\x0f\0", 2, 1, korgm); // user category

  size -= copybytes(korg9, korgm, 28);
  for (int trk = 0; trk < 6; trk++)
  {
    int prg9 = fgetc(korg9);
    int bank9 = fgetc(korg9);
    int prgm = prg9;
    int bankm = bank9;
    if (bank9 == 2)
    {
      switch(prg9)
      {
      case 1: prgm = 3; break;
      case 2: prgm = 5; break;
      case 3: prgm = 7; break;
      case 4: prgm = 8; break;
      case 5: prgm = 10; break;
      case 6: prgm = destkorgm ? 14 : 32; break;
      case 7: prgm = destkorgm ? 15 : 33; break;
      }
      bankm = 4;
    }
    fputc(prgm, korgm);
    fputc(bankm, korgm);
    size -= 2;
    size -= copybytes(korg9, korgm, 2);
  }
  size -= copybytes(korg9, korgm, 472);
//  long pos = ftell(korgm);
  for (int i = 0; i< 317; i++)
  {
     unsigned ofs = readword(korg9) + 8;
     writeword(ofs, korgm);
     size-=2;
  }
  while (size-- > 0)
   fputc(fgetc(korg9), korgm);
}

static int convertkorg9tom()
{
  korg9 = fopen(korg9name, "rb");
  if (!korg9)
  {
    perror(korg9name);
    return 0;
  }
  char id[16];
  fread(id, 1, 16, korg9);
  if (strncmp(id, "KORG9", 5) != 0)
  {
    fprintf(stderr, "%s is not a KORG9 file\n", korg9name);
    fclose(korg9);
    return 0;
  }

  // skip header
  fseek(korg9, 16, SEEK_SET);

  long stylenameofs;
  long stylenamesize;
  long styleinfoofs;
  long styleinfosize;
  stylenameofs = readlong(korg9);
  stylenamesize = readlong(korg9);
  styleinfoofs = readlong(korg9);
  styleinfosize = readlong(korg9);
  int stylecount = (int)(styleinfosize / 6);
  if (stylecount > 4)
    stylecount= 4;

  char stylename[4][10];
  fseek(korg9, stylenameofs,SEEK_SET);
  fread(stylename, 10, stylecount, korg9);

  long styleofs[4], stylesize[4];

  styleofs[0] = styleinfoofs + styleinfosize;

  fseek(korg9, styleinfoofs, SEEK_SET);
  for (int i = 0; i < stylecount; i++)
  {
    unsigned char c[6];
    fread(c, 6, 1, korg9);

    stylesize[i] = (long(c[5]) << 8) + c[4];
    if (i < stylecount-1)
      styleofs[i+1] = styleofs[i]+ stylesize[i];
  }

  korgm = fopen(korgmname, "wb");
  if (!korgm)
  {
    perror(korgmname);
    return 0;
  }
  char header[11+1] = "\x05\x04\0\0\0\0\0\0\0\0\0";

  if (destkorgm)
  {
    header[0] = 0;
    header[2] = 1;
  }

  fwrite(KorgFileID(), 5, 1, korgm);
  fwrite( header, 11, 1, korgm);
  fwrite("\xE0\0\0\0", 4, 1, korgm);

  int destcount = destkorgm ? 16 : 48;

  fputc(0, korgm);
  fputc(destcount, korgm);

  i = 0;
  while (i < stylecount)
  {
      fputc(i, korgm);
      fputc(1, korgm);
      writeword((unsigned)(stylesize[i]+8), korgm);
    i++;
  }

  while (i < destcount)
  {
      fputc(i, korgm);
      fputc(1, korgm);
      writeword(0, korgm);
    i++;
  }

  while (ftell(korgm) < 0xe0)
    fputc(0, korgm);

  for (i = 0; i < stylecount; i++)
    copystyle(stylename[i], styleofs[i], stylesize[i]);

  fclose(korgm);
  fclose(korg9);
  return 1;
}

int sty9convert(int argc,char** argv, int destformat)
{
  destkorgm = destformat;

  argc--; argv++;

  if (argc != 2)
  {
    if (destkorgm)
      printf("usage: sty9tom korgsty9.sty korgstym.sty\n");
    else
      printf("usage: sty9toi korgsty9.sty korgstyi.sty\n");
    printf("korgsty9.sty\ta korg format 9 style (e.g. for i2,i3,... keyboard)\n");

    if (destkorgm)
      printf("korgstym.sty\ta korg format M style (e.g. for is40,is50 keyboard)\n");
    else
      printf("korgstyi.sty\ta korg format I style (e.g. for i30 keyboard)\n");
    return 1;
  }
  korg9name = *argv++; argc--;
  korgmname = *argv++; argc--;

  if (convertkorg9tom())
  {
    printf("%s style file written to %s\n", KorgFileID(), korgmname);
    return 0;
  }
  return 1;
}

#ifdef NOMAIN
int sty9tomi(int argc, char** argv)
#else
int main(int argc, char** argv)
#endif
{
#ifdef KORGM
  destkorgm = 1;
#elif defined(KORGI)
  destkorgm = 0;
#else
#  error KORGM or KORGI must be defined
#endif

   return sty9convert(argc,argv, destkorgm);
}

int main_sty9tom(int argc, char** argv)
{
   return sty9convert(argc,argv, 1);
}

int main_sty9toi(int argc, char** argv)
{
   return sty9convert(argc,argv, 0);
}
